# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""This module contains the python files generated by ANTLR. These are version controlled
to allow users without a working ANTLR and/or Java setup to also use M2-ISA-R.
"""

from .CoreDSL2Lexer import CoreDSL2Lexer
from .CoreDSL2Listener import CoreDSL2Listener
from .CoreDSL2Parser import CoreDSL2Parser
from .CoreDSL2Visitor import CoreDSL2Visitor