; NOTE: Assertions have been autogenerated by utils/update_llc_test_checks.py
; RUN: llc -O0 -mtriple=riscv32 -mattr=+m,+xcorevalu -verify-machineinstrs -global-isel=1 < %s \
; RUN:   | FileCheck %s --check-prefixes=CHECK,CHECK-GISEL

define i32 @exthz(i16 %a) {
; CHECK-LABEL: exthz:
; CHECK:       # %bb.0:
; CHECK-NEXT:    # kill: def $x11 killed $x10
; CHECK-GISEL-NEXT:    seal5.cv.exthz a0, a0
; CHECK-NEXT:    ret
  %1 = zext i16 %a to i32
  ret i32 %1
}
